import json
from typing import Any, Dict

from aiohttp import BasicAuth, ClientSession, ClientConnectionError

from .const import HTTP_OK, HTTP_UNAUTHORIZED


class BaseAPI:
    """Class to perform CMI API requests"""

    def __init__(self,
                 username: str,
                 password: str,
                 session: ClientSession = None
                 ):
        """Initialize."""
        self.auth = BasicAuth(username, password)
        self.session = session

    async def _makeRequest(self, url: str) -> Dict[str, Any]:
        """Retrieve data from CMI API."""
        rawResponse: str = await self._makeRequestNoJson(url)
        data = json.loads(rawResponse)

        if data["Status code"] == 0:
            return data
        elif data["Status code"] == 1:
            raise ApiError("Node not available")
        elif data["Status code"] == 2:
            raise ApiError("Failure during the CAN-request/parameter not available for this device")
        elif data["Status code"] == 4:
            raise RateLimitError("Only one request per minute is permitted")
        elif data["Status code"] == 5:
            raise ApiError("Device not supported")
        elif data["Status code"] == 7:
            raise ApiError("CAN Bus is busy")
        else:
            raise ApiError("Unknown error")

    async def _makeRequestNoJson(self, url: str) -> str:
        """Retrieve data from CMI API that is not valid json."""
        internalSession: bool = False
        if self.session is None:
            internalSession = True
            self.session = ClientSession()

        try:
            async with self.session.get(url, auth=self.auth) as res:
                if res.status == HTTP_UNAUTHORIZED:
                    raise InvalidCredentialsError("Invalid API key")
                elif res.status != HTTP_OK:
                    raise ApiError(f"Invalid response from CMI: {res.status}")

                text = await res.text()
                if internalSession:
                    await self.session.close()
                return text
        except ClientConnectionError:
            if internalSession:
                await self.session.close()
            raise ApiError(f"Could not connect to C.M.I")

    @staticmethod
    def _is_json(toTest: str) -> bool:
        """Test if string is valid json."""
        try:
            json.loads(toTest)
        except ValueError:
            return False
        return True


class ApiError(Exception):
    """Raised when API request ended in error."""

    def __init__(self, status: str):
        """Initialize."""
        super().__init__(status)
        self.status = status


class InvalidCredentialsError(Exception):
    """Triggered when the credentials are invalid."""

    def __init__(self, status: str):
        """Initialize."""
        super().__init__(status)
        self.status = status


class RateLimitError(Exception):
    """Triggered when the rate limit is reached."""

    def __init__(self, status: str):
        """Initialize."""
        super().__init__(status)
        self.status = status



from typing import Dict, Any, List

from aiohttp import ClientSession

from .baseApi import BaseAPI
from .channel import Channel

from .const import DEVICES, ReadOnlyClass, ChannelMode


class Device(BaseAPI, metaclass=ReadOnlyClass):
    """Class to interact with a device"""

    def __init__(self, nodeID: str, host: str, username: str, password: str, session: ClientSession = None):
        """Initialize."""
        super().__init__(username, password, session)
        self.id: str = nodeID
        self.host: str = host

        self.apiVersion: int = 0
        self.deviceType: str = "00"

        self.inputs: Dict[int, Channel] = {}
        self.outputs: Dict[int, Channel] = {}

    def _extractDeviceInfo(self, json: Dict[str, Any]):
        """Extract device info from request response."""
        self.apiVersion: int = json["Header"]["Version"]
        self.deviceType: str = json["Header"]["Device"]

    @staticmethod
    def _extractChannels(mode: ChannelMode, rawChannels: List[Dict[str, Any]]) -> Dict[int, Channel]:
        """Extract channel info from data array from request."""
        listOfChannels: Dict[int, Channel] = {}
        for chRaw in rawChannels:
            ch: Channel = Channel(mode, chRaw)
            listOfChannels[ch.index] = ch

        return listOfChannels

    async def update(self):
        """Update data."""
        url: str = f"{self.host}/INCLUDE/api.cgi?jsonparam=I,O&jsonnode={self.id}"
        res: Dict[str, Any] = await self._makeRequest(url)

        self._extractDeviceInfo(res)
        self.inputs: Dict[int, Channel] = self._extractChannels(ChannelMode.INPUT, res["Data"]["Inputs"])
        self.outputs: Dict[int, Channel] = self._extractChannels(ChannelMode.OUTPUT, res["Data"]["Outputs"])

    def getDeviceType(self) -> str:
        return DEVICES.get(self.deviceType, "Unknown")

    def __repr__(self) -> str:
        return f"Node {self.id}: Type: {self.getDeviceType()}, Inputs: {len(self.inputs)}, Outputs: {len(self.outputs)}"



from typing import Dict, Any

from .const import ChannelMode, ChannelType, Languages, ReadOnlyClass, UNITS_DE, UNITS_EN


class Channel(metaclass=ReadOnlyClass):
    """Class to display a input or output"""

    def __init__(self, mode: ChannelMode, json: Dict[str, Any]):
        """Initialize and parse json to get properties"""
        self.mode: ChannelMode = mode
        self.type: ChannelType = json["AD"]
        self.index: int = json["Number"]
        self.value: float = json["Value"]["Value"]
        self.unit: str = json["Value"]["Unit"]

    def getUnit(self, language: Languages = Languages.EN) -> str:
        if language == Languages.EN:
            return UNITS_EN.get(self.unit, "Unknown")
        else:
            return UNITS_DE.get(self.unit, "Unknown")

    def __repr__(self) -> str:
        return f"Channel {self.index}: Type: {self.type}, Mode: {self.mode}, Value: {self.value} {self.getUnit()}"



__version__ = "0.1.3"
__author__ = 'DeerMaximum'

from .cmi import CMI
from .device import Device
from .channel import Channel
from .baseApi import ApiError, InvalidCredentialsError, RateLimitError
from .const import ChannelMode, ChannelType, Languages



from typing import List

from aiohttp import ClientSession

from .baseApi import BaseAPI
from .device import Device, ReadOnlyClass


class CMI(BaseAPI, metaclass=ReadOnlyClass):
    """Main class to interact with CMI"""

    def __init__(self, host: str, username: str, password: str, session: ClientSession = None):
        """Initialize."""
        super().__init__(username, password, session)
        self.host = host
        self.username = username
        self.password = password

    async def getDevices(self) -> List[Device]:
        """List connected devices."""
        url: str = f"{self.host}/INCLUDE/can_nodes.cgi?_=1"
        data: str = await self._makeRequestNoJson(url)

        nodeIDs: List[str] = data.split(";")

        devices: List[Device] = []

        for nID in nodeIDs:
            if len(nID) == 0:
                continue
            devices.append(Device(nID, self.host, self.username, self.password, self.session))

        return devices
